@@ -33,21 +33,35 @@ typedef struct _ParserState {
 	Py_UNICODE *index;
 } ParserState;
 
+typedef enum
+{
+	ARRAY_EMPTY,
+	ARRAY_NEED_VALUE,
+	ARRAY_GOT_VALUE
+} ParseArrayState;
+
+typedef enum
+{
+	OBJECT_EMPTY,
+	OBJECT_NEED_KEY,
+	OBJECT_GOT_VALUE
+} ParseObjectState;
+
 static PyObject *read_keyword (ParserState *state);
 static PyObject *read_string (ParserState *state);
 static PyObject *read_number (ParserState *state);
 static PyObject *read_array (ParserState *state);
 static PyObject *read_object (ParserState *state);
-static PyObject *_read (ParserState *state);
+static PyObject *json_read (ParserState *state);
 
-static PyObject *JSON_ReadError;
-static PyObject *JSON_LeadingZeroError;
-static PyObject *JSON_BadObjectKeyError;
-static PyObject *JSON_MissingSurrogateError;
+static PyObject *ReadError;
+static PyObject *LeadingZeroError;
+static PyObject *BadObjectKeyError;
+static PyObject *MissingSurrogateError;
 static PyObject *_Decimal;
 
 static void
-skipSpaces (ParserState *state)
+skip_spaces (ParserState *state)
 {
 	while (state->index && Py_UNICODE_ISSPACE (*state->index))
 		state->index++;
@@ -85,8 +99,7 @@ unicode_utf8_strncmp (Py_UNICODE *unicode, const char *utf8, Py_ssize_t count)
 static PyObject *
 keyword_compare (ParserState *state, const char *expected, PyObject *retval)
 {
-	ptrdiff_t left;
-	size_t len = strlen (expected);
+	size_t left, len = strlen (expected);
 	
 	left = state->end - state->index;
 	
@@ -112,7 +125,7 @@ read_keyword (ParserState *state)
 	if ((retval = keyword_compare (state, "false", Py_False)))
 		return retval;
 	
-	PyErr_Format (JSON_ReadError, "cannot parse JSON description");
+	PyErr_Format (ReadError, "cannot parse JSON description");
 	return NULL;
 }
 
@@ -124,7 +137,7 @@ read_4hex (Py_UNICODE *start, Py_UNICODE *retval_ptr)
 	py_long = PyLong_FromUnicode (start, 4, 16);
 	if (!py_long) return FALSE;
 	
-	(*retval_ptr) = PyLong_AsUnsignedLong (py_long);
+	(*retval_ptr) = (Py_UNICODE) (PyLong_AsUnsignedLong (py_long));
 	Py_DECREF (py_long);
 	return TRUE;
 }
@@ -143,9 +156,9 @@ read_unicode_escape (ParserState *state, Py_UNICODE *string_start,
 	
 	if (remaining < 4)
 	{
-		PyErr_Format (JSON_ReadError,
+		PyErr_Format (ReadError,
 		              "Unterminated escape in string starting at position " PY_SSIZE_T_F,
-		              (Py_ssize_t)(state->index - state->start));
+		              (Py_ssize_t) (state->index - state->start));
 		return FALSE;
 	}
 	
@@ -161,18 +174,18 @@ read_unicode_escape (ParserState *state, Py_UNICODE *string_start,
 		
 		if (remaining < 10)
 		{
-			PyErr_Format (JSON_MissingSurrogateError,
+			PyErr_Format (MissingSurrogateError,
 			              "Surrogate pair half is required at " PY_SSIZE_T_F,
-			              (Py_ssize_t)(state->index - state->start));
+			              (Py_ssize_t) (state->index - state->start));
 			return FALSE;
 		}
 		
 		if (string_start[(*index_ptr)] != '\\' ||
 		    string_start[(*index_ptr) + 1] != 'u')
 		{
-			PyErr_Format (JSON_MissingSurrogateError,
+			PyErr_Format (MissingSurrogateError,
 			              "Surrogate pair half is required at " PY_SSIZE_T_F,
-			              (Py_ssize_t)(state->index - state->start));
+			              (Py_ssize_t) (state->index - state->start));
 			return FALSE;
 		}
 		(*index_ptr) += 2;
@@ -221,9 +234,9 @@ read_string (ParserState *state)
 		c = start[ii];
 		if (c == 0)
 		{
-			PyErr_Format (JSON_ReadError,
+			PyErr_Format (ReadError,
 			              "unterminated string starting at position " PY_SSIZE_T_F,
-			              (Py_ssize_t)(state->index - state->start));
+			              (Py_ssize_t) (state->index - state->start));
 			return NULL;
 		}
 		
@@ -291,9 +304,9 @@ read_string (ParserState *state)
 				
 				default:
 				{
-					PyErr_Format (JSON_ReadError,
+					PyErr_Format (ReadError,
 					              "Illegal escape code '" PY_UNICODE_F "' at position " PY_SSIZE_T_F,
-					              c, (Py_ssize_t)(start - state->start) + ii);
+					              c, (Py_ssize_t) (start - state->start) + ii);
 					
 					PyMem_Free (buffer);
 					return NULL;
@@ -338,7 +351,7 @@ read_number (ParserState *state)
 	
 	while ((c = *ptr))
 	{
-		switch(c) {
+		switch (c) {
 		case '0':
 			if (!got_digit)
 			{
@@ -346,7 +359,7 @@ read_number (ParserState *state)
 			}
 			if (got_digit && leading_zero)
 			{
-				PyErr_Format (JSON_LeadingZeroError, "Invalid leading zero");
+				PyErr_Format (LeadingZeroError, "Invalid leading zero");
 				return NULL;
 			}
 			got_digit = TRUE;
@@ -362,7 +375,7 @@ read_number (ParserState *state)
 		case '9':
 			if (leading_zero)
 			{
-				PyErr_Format (JSON_LeadingZeroError, "Invalid leading zero");
+				PyErr_Format (LeadingZeroError, "Invalid leading zero");
 				return NULL;
 			}
 			got_digit = TRUE;
@@ -406,8 +419,9 @@ read_number (ParserState *state)
 	
 	if (object == NULL)
 	{
-		PyErr_Format(JSON_ReadError, "invalid number starting at position " PY_SSIZE_T_F,
-		             (Py_ssize_t)(state->index - state->start));
+		PyErr_Format (ReadError,
+		              "invalid number starting at position " PY_SSIZE_T_F,
+		              (Py_ssize_t) (state->index - state->start));
 		return NULL;
 	}
 	
@@ -415,187 +429,221 @@ read_number (ParserState *state)
 	return object;
 }
 
-
-static PyObject *
-read_array(ParserState *jsondata)
+static int
+read_array_impl (PyObject *list, ParserState *state)
 {
-	PyObject *object, *item;
-	int expect_item, items, result;
 	Py_UNICODE *start, c;
+	ParseArrayState array_state = ARRAY_EMPTY;
 	
-	object = PyList_New(0);
-	
-	start = jsondata->index;
-	jsondata->index++;
-	expect_item = TRUE;
-	items = 0;
-	while (TRUE) {
-		skipSpaces(jsondata);
-		c = *jsondata->index;
-		if (c == 0) {
-			PyErr_Format(JSON_ReadError,
-			             "unterminated array starting at "
-			             "position " PY_SSIZE_T_F,
-			             (Py_ssize_t)(start - jsondata->start));
-			goto failure;;
-		} else if (c == ']') {
-			if (expect_item && items>0) {
-				PyErr_Format(JSON_ReadError,
-				             "expecting array item at "
-				             "position " PY_SSIZE_T_F,
-				             (Py_ssize_t)(jsondata->index - jsondata->start));
-				goto failure;
+	start = state->index;
+	state->index++;
+	while (TRUE)
+	{
+		skip_spaces (state);
+		c = *state->index;
+		if (c == 0)
+		{
+			PyErr_Format (ReadError,
+			              "Unterminated array starting at "
+			              "position " PY_SSIZE_T_F,
+			              (Py_ssize_t) (start - state->start));
+			return FALSE;
+		}
+		
+		else if (c == ']')
+		{
+			if (array_state == ARRAY_NEED_VALUE)
+			{
+				PyErr_Format (ReadError,
+				              "Expecting array item at "
+				              "position " PY_SSIZE_T_F,
+				              (Py_ssize_t) (state->index - state->start));
+				return FALSE;
 			}
-			jsondata->index++;
-			break;
-		} else if (c == ',') {
-			if (expect_item) {
-				PyErr_Format(JSON_ReadError,
-				             "expecting array item at "
-				             "position " PY_SSIZE_T_F,
-				             (Py_ssize_t)(jsondata->index - jsondata->start));
-				goto failure;
+			state->index++;
+			return TRUE;
+		}
+		
+		else if (c == ',')
+		{
+			if (array_state != ARRAY_GOT_VALUE)
+			{
+				PyErr_Format (ReadError,
+				              "Expecting array item at "
+				              "position " PY_SSIZE_T_F,
+				              (Py_ssize_t) (state->index - state->start));
+				return FALSE;
 			}
-			expect_item = TRUE;
-			jsondata->index++;
-			continue;
-		} else {
-			item = _read (jsondata);
-			if (item == NULL)
-				goto failure;
-			result = PyList_Append(object, item);
-			Py_DECREF(item);
-			if (result == -1)
-				goto failure;
-			expect_item = FALSE;
-			items++;
+			array_state = ARRAY_NEED_VALUE;
+			state->index++;
+		}
+		
+		else
+		{
+			PyObject *value;
+			if ((value = json_read (state)))
+			{
+				int result = PyList_Append (list, value);
+				Py_DECREF (value);
+				if (result != -1)
+				{
+					array_state = ARRAY_GOT_VALUE;
+					continue;
+				}
+			}
+			return FALSE;
 		}
 	}
-
-	return object;
-
-failure:
-	Py_DECREF(object);
-	return NULL;
 }
 
-
 static PyObject *
-read_object(ParserState *jsondata)
+read_array (ParserState *state)
+{
+	PyObject *object = PyList_New (0);
+	
+	if (!read_array_impl (object, state))
+	{
+		Py_DECREF (object);
+		return NULL;
+	}
+	
+	return object;
+}
+
+static int
+read_object_impl (PyObject *object, ParserState *state)
 {
-	PyObject *object, *key, *value;
-	int expect_key, items, result;
 	Py_UNICODE *start, c;
+	ParseObjectState object_state = OBJECT_EMPTY;
 	
-	object = PyDict_New ();
-	
-	expect_key = TRUE;
-	items = 0;
-	start = jsondata->index;
-	jsondata->index++;
-	
-	while (TRUE) {
-		skipSpaces (jsondata);
-		c = *jsondata->index;
-		if (c == 0) {
-			PyErr_Format(JSON_ReadError,
-			             "unterminated object starting at "
-			             "position " PY_SSIZE_T_F,
-			             (Py_ssize_t)(start - jsondata->start));
-			goto failure;;
-		} else if (c == '}') {
-			if (expect_key && items>0) {
-				PyErr_Format(JSON_ReadError,
-				             "expecting object property name"
-				             " at position " PY_SSIZE_T_F,
-				             (Py_ssize_t)(jsondata->index - jsondata->start));
-				goto failure;
+	start = state->index;
+	state->index++;
+	while (TRUE)
+	{
+		skip_spaces (state);
+		c = *state->index;
+		if (c == 0)
+		{
+			PyErr_Format (ReadError,
+			              "Unterminated object starting at "
+			              "position " PY_SSIZE_T_F,
+			              (Py_ssize_t) (start - state->start));
+			return FALSE;
+		}
+		
+		else if (c == '}')
+		{
+			if (object_state == OBJECT_NEED_KEY)
+			{
+				PyErr_Format (ReadError,
+				              "Expecting object property name "
+				              "at position " PY_SSIZE_T_F,
+				              (Py_ssize_t) (state->index - state->start));
+				return FALSE;
 			}
-			jsondata->index++;
-			break;
-		} else if (c == ',') {
-			if (expect_key) {
-				PyErr_Format(JSON_ReadError,
-				             "expecting object property name"
-				             "at position " PY_SSIZE_T_F,
-				             (Py_ssize_t)(jsondata->index - jsondata->start));
-				goto failure;
-			}
-			expect_key = TRUE;
-			jsondata->index++;
-			continue;
-		} else {
-			if (c != '"') {
-				PyErr_Format(JSON_BadObjectKeyError,
-				             "expecting property name in "
-				             "object at position " PY_SSIZE_T_F,
-				             (Py_ssize_t)(jsondata->index - jsondata->start));
-				goto failure;
+			state->index++;
+			return TRUE;
+		}
+		
+		else if (c == ',')
+		{
+			if (object_state != OBJECT_GOT_VALUE)
+			{
+				PyErr_Format (ReadError,
+				              "Expecting object property name "
+				              "at position " PY_SSIZE_T_F,
+				              (Py_ssize_t) (state->index - state->start));
+				return FALSE;
 			}
+			object_state = OBJECT_NEED_KEY;
+			state->index++;
+		}
+		
+		else if (c == '"')
+		{
+			PyObject *key, *value;
+			int result;
 
-			key = _read (jsondata);
-			if (key == NULL)
-				goto failure;
-
-			skipSpaces(jsondata);
-			if (*jsondata->index != ':') {
-				PyErr_Format(JSON_ReadError,
-				             "missing colon after object "
-				             "property name at position " PY_SSIZE_T_F,
-				             (Py_ssize_t)(jsondata->index - jsondata->start));
-				Py_DECREF(key);
-				goto failure;
-			} else {
-				jsondata->index++;
+			if (!(key = json_read (state)))
+				return FALSE;
+			
+			skip_spaces (state);
+			if (*state->index != ':')
+			{
+				PyErr_Format (ReadError,
+				              "Missing colon after object "
+				              "property name at position " PY_SSIZE_T_F,
+				              (Py_ssize_t) (state->index - state->start));
+				Py_DECREF (key);
+				return FALSE;
 			}
-
-			value = _read (jsondata);
-			if (value == NULL) {
-				Py_DECREF(key);
-				goto failure;
+			
+			state->index++;
+			if (!(value = json_read (state)))
+			{
+				Py_DECREF (key);
+				return FALSE;
 			}
-
-			result = PyDict_SetItem(object, key, value);
-			Py_DECREF(key);
-			Py_DECREF(value);
+			
+			result = PyDict_SetItem (object, key, value);
+			Py_DECREF (key);
+			Py_DECREF (value);
 			if (result == -1)
-				goto failure;
-			expect_key = FALSE;
-			items++;
+			{
+				return FALSE;
+			}
+			
+			object_state = OBJECT_GOT_VALUE;
+		}
+		
+		else
+		{
+			PyErr_Format (BadObjectKeyError,
+			              "Expecting property name in "
+			              "object at position " PY_SSIZE_T_F,
+			              (Py_ssize_t) (state->index - state->start));
+			return FALSE;
 		}
 	}
-
-	return object;
-
-failure:
-	Py_DECREF(object);
-	return NULL;
 }
 
+static PyObject *
+read_object (ParserState *state)
+{
+	PyObject *object = PyDict_New ();
+	
+	if (!read_object_impl (object, state))
+	{
+		Py_DECREF (object);
+		return NULL;
+	}
+	
+	return object;
+}
 
 static PyObject *
-_read (ParserState *jsondata)
+json_read (ParserState *state)
 {
 	PyObject *object;
 
-	skipSpaces (jsondata);
-	switch (*jsondata->index) {
+	skip_spaces (state);
+	switch (*state->index) {
 	case 0:
-		PyErr_SetString (JSON_ReadError, "empty JSON description");
+		PyErr_SetString (ReadError, "Empty JSON expression");
 		return NULL;
 	case '{':
-		object = read_object (jsondata);
+		object = read_object (state);
 		break;
 	case '[':
-		object = read_array (jsondata);
+		object = read_array (state);
 		break;
 	case '"':
-		object = read_string (jsondata);
+		object = read_string (state);
 		break;
 	case 't':
 	case 'f':
 	case 'n':
-		object = read_keyword (jsondata);
+		object = read_keyword (state);
 		break;
 	case '-':
 	case '0':
@@ -608,10 +656,10 @@ _read (ParserState *jsondata)
 	case '7':
 	case '8':
 	case '9':
-		object = read_number (jsondata);
+		object = read_number (state);
 		break;
 	default:
-		PyErr_SetString (JSON_ReadError, "cannot parse JSON description");
+		PyErr_SetString (ReadError, "Cannot parse JSON expression");
 		return NULL;
 	}
 
@@ -619,14 +667,14 @@ _read (ParserState *jsondata)
 }
 
 static PyObject*
-JSON_decode(PyObject *self, PyObject *args, PyObject *kwargs)
+_read_entry (PyObject *self, PyObject *args, PyObject *kwargs)
 {
-	static char *kwlist[] = {"json", NULL};
+	static char *kwlist[] = {"string", NULL};
 	PyObject *result, *unicode;
 	ParserState state;
 	
-	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O:read", kwlist,
-	                                 &unicode))
+	if (!PyArg_ParseTupleAndKeywords (args, kwargs, "O:_read", kwlist,
+	                                  &unicode))
 		return NULL;
 	
 	Py_INCREF (unicode);
@@ -635,15 +683,15 @@ JSON_decode(PyObject *self, PyObject *args, PyObject *kwargs)
 	state.end = state.start + PyUnicode_GetSize (unicode);
 	state.index = state.start;
 	
-	result = _read (&state);
-	
-	if (result != NULL) {
-		skipSpaces (&state);
-		if (state.index < state.end) {
-			PyErr_Format(JSON_ReadError,
-			             "extra data after JSON description"
-			             " at position " PY_SSIZE_T_F,
-			             (Py_ssize_t)(state.index - state.start));
+	if ((result = json_read (&state)))
+	{
+		skip_spaces (&state);
+		if (state.index < state.end)
+		{
+			PyErr_Format (ReadError,
+			              "Extra data after JSON expression "
+			              "at position " PY_SSIZE_T_F,
+			              (Py_ssize_t) (state.index - state.start));
 			Py_DECREF (result);
 			result = NULL;
 		}
@@ -654,19 +702,16 @@ JSON_decode(PyObject *self, PyObject *args, PyObject *kwargs)
 	return result;
 }
 
-
-/* List of functions defined in the module */
-
 static PyMethodDef reader_methods[] = {
-	{"_read", (PyCFunction)JSON_decode, METH_VARARGS|METH_KEYWORDS,
-	PyDoc_STR("_read (string) -> parse the JSON representation into\n"
-	          "python objects.")},
+	{"_read", (PyCFunction) (_read_entry), METH_VARARGS|METH_KEYWORDS,
+	PyDoc_STR ("_read (string) -> Deserialize the JSON expression to\n"
+	           "a Python object.")},
 	
 	{NULL, NULL}
 };
 
-PyDoc_STRVAR(module_doc,
-"Fast JSON encoder/decoder module."
+PyDoc_STRVAR (module_doc,
+	"Fast implementation of jsonlib._read."
 );
 
 PyMODINIT_FUNC
@@ -674,20 +719,17 @@ init_reader (void)
 {
 	PyObject *m, *errors, *decimal_module;
 	
-	m = Py_InitModule3("_reader", reader_methods, module_doc);
-	
-	if (m == NULL)
+	if (!(m = Py_InitModule3 ("_reader", reader_methods, module_doc)))
 		return;
-	
 	if (!(errors = PyImport_ImportModule ("errors")))
 		return;
-	if (!(JSON_ReadError = PyObject_GetAttrString (errors, "ReadError")))
+	if (!(ReadError = PyObject_GetAttrString (errors, "ReadError")))
 		return;
-	if (!(JSON_LeadingZeroError = PyObject_GetAttrString (errors, "LeadingZeroError")))
+	if (!(LeadingZeroError = PyObject_GetAttrString (errors, "LeadingZeroError")))
 		return;
-	if (!(JSON_BadObjectKeyError = PyObject_GetAttrString (errors, "BadObjectKeyError")))
+	if (!(BadObjectKeyError = PyObject_GetAttrString (errors, "BadObjectKeyError")))
 		return;
-	if (!(JSON_MissingSurrogateError = PyObject_GetAttrString (errors, "MissingSurrogateError")))
+	if (!(MissingSurrogateError = PyObject_GetAttrString (errors, "MissingSurrogateError")))
 		return;
 	
 	if (!(decimal_module = PyImport_ImportModule ("decimal")))
